import csv
import datetime
import logging
import codecs

from .subject_import_reader import SubjectImportReader
from ..models import StudySubject, Subject, Study
from ..models.constants import GLOBAL_STUDY_ID

CSV_DATE_FORMAT = "%d/%m/%Y"

logger = logging.getLogger(__name__)


class TnsCsvSubjectImportReader(SubjectImportReader):
    def __init__(self):
        self.study = Study.objects.filter(id=GLOBAL_STUDY_ID)[0]

    def load_data(self, filename):
        study_subjects = []
        with open(filename) as csv_file:
            reader = csv.reader((remove_bom(line) for line in csv_file), delimiter=';')
            headers = next(reader, None)
            for row in reader:
                subject = Subject()
                study_subject = StudySubject()
                study_subject.subject = subject
                study_subject.study = self.study
                for header, value in zip(headers, row):
                    self.add_data(study_subject, header, value)
                if study_subject.nd_number is None or study_subject.nd_number == "":
                    study_subject.nd_number = study_subject.screening_number
                study_subjects.append(study_subject)
        return study_subjects

    def add_data(self, study_subject, column_name, value):
        # type: (StudySubject, str, str) ->  None
        if column_name == "firstname" or column_name == "sig_firstname":
            study_subject.subject.first_name = self.get_new_value(study_subject.subject.first_name, column_name, value)
        elif column_name == "lastname" or column_name == "sig_lastname":
            study_subject.subject.last_name = self.get_new_value(study_subject.subject.last_name, column_name, value)
        elif column_name == "donor_id":
            study_subject.screening_number = self.get_new_value(study_subject.screening_number, column_name, value)
        elif column_name == "email":
            study_subject.subject.email = self.get_new_value(study_subject.subject.email, column_name, value)
        elif column_name == "address":
            study_subject.subject.address = self.get_new_value(study_subject.subject.address, column_name, value)
        elif column_name == "postal_code":
            study_subject.subject.postal_code = self.get_new_value(study_subject.subject.postal_code, column_name,
                                                                   value)
        elif column_name == "city":
            study_subject.subject.city = self.get_new_value(study_subject.subject.city, column_name, value)
        elif column_name == "phonenr":
            study_subject.subject.phone_number = self.get_new_value(study_subject.subject.phone_number, column_name,
                                                                    value)
        elif column_name == "representative":
            study_subject.subject.next_of_keen_name = self.get_new_value(study_subject.subject.next_of_keen_name,
                                                                         column_name, value)
        elif column_name == "treatingphysician":
            if value is not None and value != "":
                value = "Treating physician: " + value
            study_subject.comments = self.get_new_value(study_subject.comments, column_name, value)
        elif column_name == "dateofbirth":
            study_subject.subject.date_born = self.get_new_date_value(study_subject.subject.date_born, column_name,
                                                                      value)
        else:
            logger.warn("Don't know how to handle column " + column_name + " with data " + value)

    def get_new_value(self, old_value, column_name, new_value):
        # type: (unicode,unicode,unicode) ->  unicode
        if old_value is None or old_value == "":
            return new_value
        if new_value is None or new_value == "":
            return old_value
        if old_value == new_value:
            return new_value
        logger.warn(
            "Contradicting entries in csv file for column: " + column_name + "(" + new_value + "," + old_value +
            "). Latest value will be used")
        return new_value

    def get_new_date_value(self, old_value, column_name, new_value):
        # type: (datetime,unicode,unicode) ->  datetime
        if old_value is None or old_value == "":
            try:
                result = datetime.datetime.strptime(new_value, CSV_DATE_FORMAT)
            except ValueError:
                logger.warn("Invalid date: " + new_value)
                result = old_value
            return result
        if new_value is None or new_value == "":
            return old_value
        logger.warn(
            "Contradicting entries in csv file for column: " + column_name + "(" + new_value + "," + old_value +
            "). Latest value will be used")
        return datetime.datetime.strptime(new_value, CSV_DATE_FORMAT)


def remove_bom(line):
    return line[3:] if line.startswith(codecs.BOM_UTF8) else line
