# coding=utf-8
import codecs
import csv
import datetime
import logging
import sys
import traceback

import pytz
from django.conf import settings

from .warning_counter import MsgCounterHandler
from web.models import StudySubject, Study, Visit, Appointment, AppointmentType, Location, AppointmentTypeLink, Subject, User, Worker, Provenance
from web.models.constants import GLOBAL_STUDY_ID

CSV_DATE_FORMAT = "%d/%m/%Y"

logger = logging.getLogger(__name__)


class TnsCsvVisitImportReader:
    def __init__(self):
        self.study = Study.objects.filter(id=GLOBAL_STUDY_ID)[0]
        appointment_code = getattr(settings, "IMPORT_APPOINTMENT_TYPE", "SAMPLES")

        appointment_types = AppointmentType.objects.filter(code=appointment_code)
        if len(appointment_types) > 0:
            self.appointment_type = appointment_types[0]
        else:
            logger.warn("Appointment type does not exist: " + appointment_code)
            self.appointment_type = None
        self.problematic_count = 0
        self.processed_count = 0
        self.warning_count = 0

        self.importer_user = None

        importer_user_name = getattr(settings, "IMPORTER_USER", None)
        if importer_user_name is not None:
            user = User.objects.filter(username=importer_user_name)
            if user is None:
                logger.warn("User does not exist: " + importer_user_name)
            else:
                self.importer_user = Worker.objects.filter(user=user)

    def load_data(self, filename):
        warning_counter = MsgCounterHandler()
        logging.getLogger('').addHandler(warning_counter)

        result = []
        with open(filename) as csv_file:
            reader = csv.reader((remove_bom(line) for line in csv_file), delimiter=';')
            headers = next(reader, None)
            for row in reader:
                try:
                    data = {}
                    for h, v in zip(headers, row):
                        data[h] = v
                    nd_number = data['donor_id']
                    study_subjects = StudySubject.objects.filter(nd_number=nd_number)
                    if len(study_subjects) == 0:
                        logger.debug("Subject " + nd_number + " does not exist. Creating")
                        subject = Subject.objects.create()
                        study_subject = StudySubject.objects.create(subject=subject,
                                                                    study=Study.objects.filter(id=GLOBAL_STUDY_ID)[0],
                                                                    nd_number=nd_number,
                                                                    screening_number=nd_number)
                    else:
                        study_subject = study_subjects[0]
                    date = self.extract_date(data['dateofvisit'])

                    location = self.extract_location(data)

                    visit_number = data['visit_id']
                    visit_number = int(visit_number) + 1

                    for i in range(1, visit_number, 1):
                        if Visit.objects.filter(subject=study_subject, visit_number=i).count() == 0:
                            logger.warn(
                                "Previous visit (visit " + str(
                                    i) + ") for subject " + nd_number + " does not exist. Creating empty")
                            Visit.objects.create(subject=study_subject, visit_number=i,
                                                 datetime_begin=date + datetime.timedelta(days=-1, minutes=i),
                                                 datetime_end=date + datetime.timedelta(days=-1, minutes=i))
                        else:
                            break

                    visits = Visit.objects.filter(subject=study_subject, visit_number=visit_number)

                    if len(visits) > 0:
                        logger.debug("Visit for subject " + nd_number + " already exists. Updating")
                        visit = visits[0]

                        changes = [('datetime_begin', date),
                                   ('datetime_end', date + datetime.timedelta(days=14))]

                        for field, new_value in changes:
                            old_value = getattr(visit, field)
                            if old_value == new_value:
                                continue
                            description = '{} changed from "{}" to "{}"'.format(field, old_value, new_value)
                            p = Provenance(modified_table=Visit._meta.db_table,
                                        modified_table_id=visit.id,
                                        modification_author=self.importer_user,
                                        previous_value=old_value,
                                        new_value=new_value,
                                        modification_description=description,
                                        modified_field=field,
                                        )
                            setattr(visit, field, new_value)
                            p.save()
                        visit.save()
                    else:
                        visit = Visit.objects.create(subject=study_subject, visit_number=visit_number,
                                                     datetime_begin=date,
                                                     datetime_end=date + datetime.timedelta(days=14))
                        visit.save()
                        #visit does not have id until .save() is done
                        for field in Visit._meta.get_fields():
                            if field.get_internal_type() == "CharField" or field.get_internal_type() == "DateField" or field.get_internal_type() is "BooleanField":
                                new_value = getattr(visit, field.name)
                                if new_value is not None and new_value != "":
                                    description = '{} changed from "{}" to "{}"'.format(field, '', new_value)
                                    p = Provenance(modified_table=Visit._meta.db_table,
                                                modified_table_id=visit.id,
                                                modification_author=self.importer_user,
                                                previous_value='',
                                                new_value=new_value,
                                                modification_description=description,
                                                modified_field=field,
                                                )
                                    p.save()
                        
                    result.append(visit)

                    appointments = Appointment.objects.filter(visit=visit, appointment_types=self.appointment_type)
                    if len(appointments) > 0:
                        logger.debug("Appointment for subject " + nd_number + " already set. Updating")
                        appointment = appointments[0]
                        
                        #(field, new_value)
                        changes = [('length', 60),
                                   ('datetime_when', date),
                                   ('location', location)]

                        for field, new_value in changes:
                            old_value = getattr(appointment, field)
                            if old_value == new_value:
                                continue
                            description = '{} changed from "{}" to "{}"'.format(field, old_value, new_value)
                            p = Provenance(modified_table=Appointment._meta.db_table,
                                        modified_table_id=appointment.id,
                                        modification_author=self.importer_user,
                                        previous_value=old_value,
                                        new_value=new_value,
                                        modification_description=description,
                                        modified_field=field,
                                        )
                            setattr(appointment, field, new_value)
                            p.save()
                        appointment.save()
                    else:
                        appointment = Appointment.objects.create(visit=visit, length=60, datetime_when=date,
                                                                 location=location)
                        if self.appointment_type is not None:
                            AppointmentTypeLink.objects.create(appointment_id=appointment.id,
                                                               appointment_type=self.appointment_type)
                        
                        appointment.save()
                        #appointment does not have id until .save() is done
                        for field in Appointment._meta.get_fields():
                            if field.get_internal_type() == "CharField" or field.get_internal_type() == "DateField" or field.get_internal_type() is "BooleanField":
                                new_value = getattr(appointment, field.name)
                                if new_value is not None and new_value != "":
                                    description = '{} changed from "{}" to "{}"'.format(field, '', new_value)
                                    p = Provenance(modified_table=Appointment._meta.db_table,
                                                modified_table_id=appointment.id,
                                                modification_author=self.importer_user,
                                                previous_value='',
                                                new_value=new_value,
                                                modification_description=description,
                                                modified_field=field,
                                                )
                                    p.save()
                    self.processed_count += 1
                except:
                    self.problematic_count += 1
                    traceback.print_exc(file=sys.stdout)
                    logger.warn("Problematic data: " + ';'.join(row))

        if "WARNING" in warning_counter.level2count:
            self.warning_count = warning_counter.level2count["WARNING"]
        logging.getLogger('').removeHandler(warning_counter)

        return result

    def extract_date(self, text):
        # type: (unicode) ->  datetime

        # by default use day after tomorrow
        result = datetime.datetime.now().replace(hour=9, minute=0) + datetime.timedelta(days=2)
        try:
            year = int(text[:4])
            month = int(text[4:6])
            day = int(text[6:8])
            result = result.replace(year=year, month=month, day=day, tzinfo=pytz.UTC)
        except ValueError:
            logger.warn("Invalid date: " + text)
        return result

    def extract_location(self, data):
        # type: (dict) ->  Location

        text = data.get('adressofvisit', None)
        if text is None:
            text = data['lab_id']

            if text == 'lab-reunis-1':
                text = 'Laboratoires réunis, 23 Route de Diekirch, 6555, Bollendorf-Pont'
            if text == 'lab-reunis-2':
                text = 'Laboratoires réunis, 38 Rue Hiehl, 6131, Junglinster'
            if text == 'lab-reunis-3':
                text = 'Laboratoires réunis, 16 Rue de la Gare, 6117, Junglinster'
            if text == 'lab-reunis-4':
                text = 'Laboratoires réunis, 456 Rue de Neudorf, 2222, Luxembourg'
            if text == 'lab-reunis-5':
                text = 'Laboratoires réunis, 14 Place St Michel, 7556, Mersch'
            if text == 'lab-reunis-6':
                text = 'Laboratoires réunis, 2 Avenue des Bains, 5610, Mondorf-les-Bains'
            if text == 'lab-reunis-7':
                text = 'Laboratoires réunis, 239 Route d\'Arlon, 8011, Strassen'
            if text == 'lab-reunis-8':
                text = 'Laboratoires réunis, 123 Route de Diekirch, 7220, Walferdange'
            if text == 'lab-reunis-9':
                text = 'Laboratoires réunis, 20 Rue de Luxembourg, 4220, Esch-sur-Alzette'
            if text == 'lab-reunis-10':
                text = 'Laboratoires réunis, 124 Avenue de Luxembourg, 4940, Bascharage'
            if text == 'lab-reunis-11':
                text = 'Laboratoires réunis, 1 Marbuergerstrooss, 9764, Marnach'
            if text == 'lab-reunis-12':
                text = 'Laboratoires réunis, 51 Avenue Lucien Salentiny, 9080, Ettelbruck'
            if text == 'lab-reunis-13':
                text = 'Laboratoires réunis, 14 route de l\'Europe, 5531, Remich'
            if text == 'lab-reunis-14':
                text = 'Laboratoires réunis, 27, rue Principale, 5240, Sandweiler'
            if text == 'lab-reunis-15':
                text = 'Laboratoires réunis, booking by phone 780 290-1, , '
            if text == 'lab-bionext-16':
                text = 'BioneXt, 2 Rue du Chateau d\'Eau, 3364, Leudelange'
            if text == 'lab-bionextpd-17':
                text = 'BioneXt, PickenDoheem'
            if text == 'lab-ketterthill-18':
                text = 'Ketterthill, 11, rue Schwaarze Wee , 3474, Dudelange'
            if text == 'lab-ketterthill-19':
                text = 'Ketterthill, 52, bd J.-F. Kennedy , 4170, Esch-sur-Alzette'
            if text == 'lab-ketterthill-20':
                text = 'Ketterthill, 7, route de Bettembourg , 5810, Hesperange'
            if text == 'lab-ketterthill-21':
                text = 'Ketterthill, Avenue des Bains (Dom. Thermal) , 5601, Mondorf-les-Bains'
            if text == 'lab-ketterthill-22':
                text = 'Ketterthill, 8, avenue du Swing , 4367, Belvaux'
            if text == 'lab-ketterthill-23':
                text = 'Ketterthill, 1-3, rue de la Continentale , 4917, Bascharage'
            if text == 'lab-ketterthill-24':
                text = 'Ketterthill, 14, rue d\'Esch, 3920, Mondercange'
            if text == 'lab-ketterthill-25':
                text = 'Ketterthill, 21, rue d\'Orval , 2270, Luxembourg'
            if text == 'lab-ketterthill-26':
                text = 'Ketterthill, 24, rue Glesener , 1630, Luxembourg'
            if text == 'lab-ketterthill-27':
                text = 'Ketterthill, 36, avenue Victor Hugo , 1750, Luxembourg'
            if text == 'lab-ketterthill-28':
                text = 'Ketterthill, 70, rue de Luxembourg , 8140, Bridel'
            if text == 'lab-ketterthill-29':
                text = 'Ketterthill, 15, rue Edward Steichen , 2540, Luxembourg'
            if text == 'lab-ketterthill-30':
                text = 'Ketterthill, 29, rue Cents , 1319, Luxembourg'
            if text == 'lab-ketterthill-31':
                text = 'Ketterthill, 155, 26-30, av. J.-F. Kennedy , 9080, Ettelbruck'
            if text == 'lab-ketterthill-32':
                text = 'Ketterthill, 12, rue G.-D. Charlotte , 7520, Mersch'
            if text == 'lab-ketterthill-33':
                text = 'Ketterthill, 18, rue de la Piscine , 8508, Redange-sur-Atert'
            if text == 'lab-ketterthill-34':
                text = 'Ketterthill, 19, rue Grande-Duchesse Charlotte , 9515, Wiltz'
            if text == 'lab-bionext-35':
                text = 'BioneXt, 18 Rue Lentz, 3509, Dudelange'
            if text == 'lab-bionext-36':
                text = 'BioneXt, 54 Rue du Maréchal Foch, 1527, Luxembourg'
            if text == 'lab-reunis-37':
                text = 'Laboratoires réunis, 5, route de Trèves , 6793, Grevenmacher'
            if text == 'lab-reunis-38':
                text = 'Laboratoires réunis, 22, rue des Ardennes, 1133, Bonnevoie'
            if text == 'lab-ketterthill-39':
                text = 'Ketterthill, 38, rue Robert Schumann, 5751, Frisange'
            if text == 'lab-reunis-40':
                text = 'Laboratoires réunis, 4, rue Tony Bourg, 1278, Gasperich'
            if text == 'lab-ketterthill-41':
                text = "Ketterthill, 143-147, route d'Arlon, 8009, Strassen"
            if text == 'lab-reunis-42':
                text = 'Laboratoires réunis, 31, route de Luxembourg, 7240, Bereldange'
            if text == 'lab-reunis-43':
                text = 'Laboratoires réunis, 11, rue de Peppange, 3270, Bettembourg'
            if text == 'lab-reunis-44':
                text = 'Laboratoires réunis, 3, Suebelwee, 5685, Dalheim'
            if text == 'lab-reunis-45':
                text = 'Laboratoires réunis, 1, rue des Glacis, 1628, Luxembourg, Glacis'

        locations = Location.objects.filter(name=text)
        if len(locations) > 0:
            return locations[0]
        else:
            logger.warn("Location with name does not exist: " + text)
            return Location.objects.create(name=text)

    def get_summary(self):
        result = "<p>Number of successfully added appointments: <b>" + str(self.processed_count) + "</b></p>"
        style = ''
        if self.problematic_count > 0:
            style = ' color="red" '
        result += "<p><font " + style + ">Number of problematic appointments: <b>" + str(
            self.problematic_count) + "</b></font></p>"
        style = ''
        if self.warning_count > 0:
            style = ' color="brown" '
        result += "<p><font " + style + ">Number of raised warnings: <b>" + str(self.warning_count) + "</b></font></p>"
        return result


def remove_bom(line):
    return line[3:] if line.startswith(codecs.BOM_UTF8) else line
